# This is an example of interacting with an ERC-721 contract, using the ERC721 interface.
import os
from fireblocks_defi_sdk_py import ERC721, Chain, Web3Bridge
from fireblocks_sdk import FireblocksSDK, TRANSACTION_STATUS_COMPLETED

API_KEY = os.environ['TEST_API_KEY']
API_SECRET = os.environ['SECRET']

SDK = FireblocksSDK(API_SECRET, API_KEY)
VAULT_ID = "0"
CONTRACT_ADDRESS = ""
RECEIVER_ADDRESS = ""
TOKEN_ID = 1

if __name__ == "__main__":
    # We start by building a bridge to the Ropsten network, including initiating a ERC721 object.
    ropsten_bridge = Web3Bridge(SDK, VAULT_ID, CONTRACT_ADDRESS, Chain.ROPSTEN)
    nft_contract_bridge = ERC721(ropsten_bridge)

    # ERC721 implements all the basic capabilities mentioned in ERC721:
    # https://ethereum.org/en/developers/docs/standards/tokens/erc-721/
    # This code will show a few examples including the inherited supportsInterface from ERC165.

    # Each interface has a unique ID. If no parameter is passed onto the function, it checks for the default interface
    # of token, meaning - 721.
    interface_supported = nft_contract_bridge.supports_interface()
    if interface_supported:
        print(f"{nft_contract_bridge} - supports ERC721 interface.")

    # Next, we will validate whether we own a balance of NFT, and attempt to send a Token to another address.
    # We can use the balance_of function, and if we don't pass any arguments to it, it will use the relevant wallet
    # address to the vault ID passed to the Web3Bridge object.
    balance = nft_contract_bridge.balance_of()
    print(f"{nft_contract_bridge.wallet_address} owns {balance} tokens.")

    if balance:
        # We will now verify if the owner of the token is indeed the wallet, and then transfer it to another address.
        token_owner = nft_contract_bridge.owner_of(TOKEN_ID)
        if token_owner == nft_contract_bridge.wallet_address:
            receiver_initial_balance = nft_contract_bridge.balance_of(RECEIVER_ADDRESS)
            # safe_transfer_from returns a dictionary with a status and id.
            transaction = nft_contract_bridge.safe_transfer_from(RECEIVER_ADDRESS, TOKEN_ID)
            transaction_result = ropsten_bridge.check_tx_is_completed(transaction['id'])

            if transaction_result == TRANSACTION_STATUS_COMPLETED:
                # After the transaction has been completed we check whether the balance of the receiver has been updated
                if nft_contract_bridge.balance_of(RECEIVER_ADDRESS) == receiver_initial_balance + 1:
                    print(f"{RECEIVER_ADDRESS} has received the token successfully.")
            else:
                print("Could not post transaction.")



from . base_token import *


class CustomToken(BaseToken):
    def __init__(self, web3_bridge: Web3Bridge, custom_abi: str):
        super().__init__(web3_bridge)
        self.abi = custom_abi
        self.contract = self.web_provider.eth.contract(
            address=self.web_provider.toChecksumAddress(self.web3_bridge.external_wallet_address),
            abi=self.abi
        )



# This is an example of interacting with an ERC-1155 contract, using the ERC1155 interface.
import os
from fireblocks_defi_sdk_py import ERC1155, Chain, Web3Bridge
from fireblocks_sdk import FireblocksSDK, TRANSACTION_STATUS_COMPLETED

API_KEY = os.environ['TEST_API_KEY']
API_SECRET = os.environ['SECRET']

SDK = FireblocksSDK(API_SECRET, API_KEY)
VAULT_ID = "0"
CONTRACT = ""
RECEIVER_ADDRESSES = ""
# A list of ints representing the token IDs.
TOKEN_IDS = [1, 2]

if __name__ == "__main__":
    # We start by building a bridge to the Kovan network, including initiating a ERC1155 object.
    kovan_bridge = Web3Bridge(SDK, VAULT_ID, CONTRACT, Chain.KOVAN)
    multi_token_contract_bridge = ERC1155(kovan_bridge)

    # ERC1155 implements all the basic capabilities mentioned in ERC721:
    # https://ethereum.org/en/developers/docs/standards/tokens/erc-1155/
    # This code will show a few examples using the functions mentioned above.

    # We will check the balance of tokens 1, 2 under the vault wallet address, and then use it to send all the tokens to
    # another address.
    tokens_balance = multi_token_contract_bridge.balance_of_batch(TOKEN_IDS)
    if tokens_balance:
        # Once again, the object will by default attempt to send assets from the given wallet address, unless defined
        # otherwise. We can also add a note to the transaction in Fireblocks.
        transfer_tokens = multi_token_contract_bridge.safe_batch_transfer_from(RECEIVER_ADDRESSES, TOKEN_IDS,
                                                                               tokens_balance,
                                                                               note="Transfer to another vault.")
        transfer_result = kovan_bridge.check_tx_is_completed(transfer_tokens['id'])

        if transfer_result == TRANSACTION_STATUS_COMPLETED:
            print(f"{RECEIVER_ADDRESSES} now owns {tokens_balance} of tokens: {TOKEN_IDS}")
        else:
            print("Failed transferring tokens.")



from . base_token import *
from .. utils.helpers import ERC721_ABI


class ERC721(BaseToken):
    def __init__(self, web3_bridge: Web3Bridge):
        super().__init__(web3_bridge)
        self.abi = ERC721_ABI
        self.contract: contract = self.web_provider.eth.contract(
            address=self.web_provider.toChecksumAddress(self.web3_bridge.external_wallet_address),
            abi=self.abi
        )

    # Payables
    def approve(self, to_address: str, token_id: int, approver_address: str = "", note: str = ""):
        """
        Receives an address and a Token ID so the specified address can perform actions on it.
        :param to_address: The receiver of the token.
        :param token_id: The ID of the sent token.
        :param approver_address: The owner of the token.
        :param note: (Optional) Add a note to the transaction.
        :return: None
        """
        checked_app_adr = self.web_provider.toChecksumAddress(to_address)
        if not approver_address:
            approver_address = self.wallet_address
        address_dict = {"from": self.web_provider.toChecksumAddress(approver_address)}
        return self.submit_transaction(self.call_write_function("approve", checked_app_adr, token_id,
                                                                building_params=address_dict), note)

    def safe_transfer_from(self, to_address: str, token_id: int, from_address: str = "", data: bytes = bytearray(),
                           note: str = ""):
        """
        :param from_address: The sender of the token.
        :param to_address: The receiver of the token.
        :param token_id: The ID of the sent token.
        :param from_address: (Optional) The sender of the token, if it's not the wallet under Fireblocks.
        :param data: (Optional) Send additional data (bytes) only if required by contract.
        :param note: (Optional) Add a note to the transaction.
        :return: None
        """
        if not from_address:
            from_address = self.wallet_address
        address_dict = {"from": self.web_provider.toChecksumAddress(from_address)}
        checked_from_adr = self.web_provider.toChecksumAddress(from_address)
        checked_to_adr = self.web_provider.toChecksumAddress(to_address)
        if data:
            transaction = self.contract.functions.safeTransferFrom(
                checked_from_adr,
                checked_to_adr,
                token_id,
                data
            ).buildTransaction(address_dict)

        else:
            transaction = self.contract.functions.safeTransferFrom(
                checked_from_adr,
                checked_to_adr,
                token_id
            ).buildTransaction(address_dict)

        return self.submit_transaction(transaction, note)

    def transfer_from(self, to_address: str, token_id: int, from_address: str = "", note: str = ""):
        """
        Do note it's encouraged to use safe_transfer_from instead.
        :param from_address: The sender of the token.
        :param to_address: The receiver of the token.
        :param token_id: The ID of the sent token.
        :param from_address: (Optional) The sender of the token, if it's not the wallet under Fireblocks.
        :param note: (Optional) Add a note to the transaction.
        :return: None
        """
        if not from_address:
            from_address = self.wallet_address
        address_dict = {"from": self.web_provider.toChecksumAddress(from_address)}
        checked_from_adr = self.web_provider.toChecksumAddress(from_address)
        checked_to_adr = self.web_provider.toChecksumAddress(to_address)
        return self.submit_transaction(self.call_write_function("transferFrom",
                                                                checked_from_adr,
                                                                checked_to_adr,
                                                                token_id,
                                                                building_params=address_dict), note)

    def set_approval_for_all(self, operator_address: str, is_approved: bool, note: str = ""):
        """
        Provide an operator with approval permission or revoke them
        :param operator_address: Approved operator address
        :param is_approved: True to permit, False to revoke
        :param note: (Optional) Add a note to the transaction.
        :return: None
        """
        checked_op_adr = self.web_provider.toChecksumAddress(operator_address)
        return self.submit_transaction(self.call_write_function("setApprovalForAll", checked_op_adr, is_approved), note)

    # Views
    def supports_interface(self, interface_id: str = "0x80ac58cd") -> bool:
        """
        Checks if contract supports a certain interface.
        :param interface_id: The interface id. "0x80ac58cd" is ERC721 interface id. "0xd9b67a26" is ERC1155 interface id
        :return: True if the contract supports the interface, False otherwise
        """
        return self.call_read_function("supportsInterface", interface_id)

    def get_approved(self, token_id: int) -> str:
        """
        Gets the approved address for a token ID, or  the zero address (0x0000000000000000000000000000000000000000)
        if no address is set. Will raise an exception if token ID does not exist.
        :param token_id: a number representing the token ID
        :return: The address approved for the token.
        """
        return self.call_read_function("getApproved", token_id)

    def is_approved_for_all(self, owner_address: str, operator_address: str) -> bool:
        """
        Validates whether an operator has been approved by a stated owner.
        :param owner_address: Address of the owner.
        :param operator_address: Address of the meant to be operator.
        :return: True whether operator is approved, False otherwise
        """
        owner_checked_address = self.web_provider.toChecksumAddress(owner_address)
        operator_checked_address = self.web_provider.toChecksumAddress(operator_address)
        return self.call_read_function("isApprovedForAll", owner_checked_address, operator_checked_address)

    def balance_of(self, owner_address: str = "") -> int:
        """
        Gets the balance of the address provided
        :param owner_address: Address to be checked
        :return: Balance (int)
        """
        if not owner_address:
            owner_address = self.wallet_address
        owner_checked_address = self.web_provider.toChecksumAddress(owner_address)
        return self.call_read_function("balanceOf", owner_checked_address)

    def owner_of(self, token_id: int) -> str:
        """
        Checks who owns the token ID.
        :param token_id: a number representing the token ID
        :return: The address (plain) of the owner
        """
        return self.call_read_function("ownerOf", token_id)

    # Events



from ... web3_bridge import Web3Bridge, CHAIN_TO_ASSET_ID
from .. utils.helpers import ADDRESS
from web3 import contract


class BaseToken:
    def __init__(self, web3_bridge: Web3Bridge):
        """
        :param web3_bridge: Web3Bridge that holds the SDK, contract address, vault ID and chain (with white listed
        address optional as well).
        """
        self.web3_bridge = web3_bridge
        self.wallet_address = self.web3_bridge.fb_api_client.get_deposit_addresses(self.web3_bridge.source_vault_id,
                                                                                   self.web3_bridge.asset)[0][ADDRESS]
        self.web_provider = web3_bridge.web_provider
        self.abi = None
        self.contract: contract = self.web_provider.eth.contract(
            address=self.web_provider.toChecksumAddress(self.web3_bridge.external_wallet_address)
        )

    def __str__(self):
        return f"Contract [{self.web3_bridge.external_wallet_address}] | Chain [{self.web3_bridge.chain.value}]"

    def call_read_function(self, abi_function: str, *args):
        """
        Use this to execute a function that needs no signing, e.g. supportsInterface. Basically a "read" function.
        :param abi_function:
        :param args:
        :return:
        """
        return self.contract.get_function_by_name(abi_function)(*args).call()

    def call_write_function(self, abi_function: str, *args, building_params: dict = None) -> dict:
        """
        Use this to build a transaction which will then be sent to fireblocks to get signed.
        :param abi_function:
        :param args:
        :param building_params:
        :return:
        """
        if not building_params:
            building_params = {}
        return self.contract.get_function_by_name(abi_function)(*args).buildTransaction(building_params)

    def submit_transaction(self, transaction: dict, note: str = "") -> dict:
        """
        Takes a ready transaction after being built (using web3 buildTransaction()) and transmits it to Fireblocks.
        :param note:
        :param transaction:
        :return:
        """
        return self.web3_bridge.send_transaction(transaction, note)



import time

from fireblocks_sdk import FireblocksSDK, TransferPeerPath, DestinationTransferPeerPath, ONE_TIME_ADDRESS, \
    VAULT_ACCOUNT, EXTERNAL_WALLET, TRANSACTION_STATUS_BLOCKED, TRANSACTION_STATUS_COMPLETED, TRANSACTION_STATUS_FAILED
from fireblocks_sdk.api_types import TRANSACTION_STATUS_CANCELLED
from chain import Chain
from web3 import Web3

SUBMIT_TIMEOUT = 45
STATUS_KEY = "status"


CHAIN_TO_ASSET_ID = {
    Chain.MAINNET: ('ETH', "https://cloudflare-eth.com"),
    Chain.ROPSTEN: ('ETH_TEST', "https://ropsten.infura.io/v3/9aa3d95b3bc440fa88ea12eaa4456161"),
    Chain.KOVAN: ('ETH_TEST2', "https://kovan.infura.io/v3/bce1459b281d491f8d06dd4f2c3d1a12"),
    Chain.BSC: ('BNB_BSC', "https://bsc-dataseed.binance.org"),
    Chain.BSC_TEST: ('BNB_TEST', "https://data-seed-prebsc-1-s3.binance.org:8545"),
    Chain.POLYGON: ('MATIC_POLYGON', "https://polygon-rpc.com")
}


class Web3Bridge:
    def __init__(self, fb_api_client: FireblocksSDK, vault_account_id: str, external_wallet_address: str, chain: Chain,
                 wl_uuid: str = ""):
        """
        :param fb_api_client: Fireblocks API client.
        :param vault_account_id: The source vault which address will sign messages.
        :param external_wallet_address: The address of the interacted *contract*.
        :param chain: Object of type Chain to represent what network to work with.
        :param wl_uuid: (Optional) If the contract is whitelisted, it can be sent through the co-responding UUID.
        """
        self.fb_api_client = fb_api_client
        self.source_vault_id = vault_account_id
        self.external_wallet_address = external_wallet_address
        self.chain = chain
        self.asset: str = CHAIN_TO_ASSET_ID[self.chain][0]
        self.web_provider = Web3(Web3.HTTPProvider(CHAIN_TO_ASSET_ID[self.chain][1]))
        self.wl_uuid = wl_uuid

    def send_transaction(self, transaction: dict, note="") -> dict:
        """
        Takes a ready transaction after being built (using web3 buildTransaction()) and transmits it to Fireblocks.
        :param transaction: A transaction object (dict) to submit to the blockchain.
        :param note: (Optional) A note to submit with the transaction.
        :return:
        """
        if self.wl_uuid:
            destination = TransferPeerPath(EXTERNAL_WALLET, self.wl_uuid)
        else:
            destination = DestinationTransferPeerPath(ONE_TIME_ADDRESS,
                                                      one_time_address={"address": self.external_wallet_address})
        return self.fb_api_client.create_transaction(
            tx_type="CONTRACT_CALL",
            asset_id=CHAIN_TO_ASSET_ID[self.chain],
            source=TransferPeerPath(VAULT_ACCOUNT, self.source_vault_id),
            amount="0",
            destination=destination,
            note=note,
            extra_parameters={
                "contractCallData": transaction["data"]
            }
        )

    def check_tx_is_completed(self, tx_id) -> dict:
        """
        This function waits for SUBMIT_TIMEOUT*4 (180 by default) seconds to retrieve status of the transaction sent to
        Fireblocks. Will stop upon completion / failure.
        :param tx_id: Transaction ID from FBKS.
        :return: Transaction last status after timeout / completion.
        """
        timeout = 0
        current_status = self.fb_api_client.get_transaction_by_id(tx_id)[STATUS_KEY]
        while current_status not in (
                TRANSACTION_STATUS_COMPLETED, TRANSACTION_STATUS_FAILED, TRANSACTION_STATUS_BLOCKED,
                TRANSACTION_STATUS_CANCELLED) and timeout < SUBMIT_TIMEOUT:
            print(f"TX [{tx_id}] is currently at status - {current_status} {'.' * (timeout % 4)}                ",
                  end="\r")
            time.sleep(4)
            current_status = self.fb_api_client.get_transaction_by_id(tx_id)[STATUS_KEY]
            timeout += 1

        print(f"\nTX [{tx_id}] is currently at status - {current_status}")
        return current_status



import requests

ADDRESS = "address"
FETCH_URL = "https://api.etherscan.io/api?module=contract&action=getabi&address="
ERROR_MESSAGES = ("Contract source code not verified",
                  "Max rate limit reached, please use API Key for higher rate limit")


def fetch_abi(contract_address: str) -> str:
    """
    Attempts to get the ABI from Etherscan through FETCH_URL + given contract address.
    :param contract_address: Address of a *verified* contract.
    :return:
    """
    fetch_url = FETCH_URL + contract_address
    resp = requests.get(fetch_url, timeout=20)
    result = resp.json()['result']
    if result == ERROR_MESSAGES[0]:
        raise LookupError("Could not find a verified contract ABI.")
    elif result == ERROR_MESSAGES[1]:
        raise OverflowError("You have sent too many requests, please make sure you're not passing the rate limit of 1"
                            " request per 1.5 seconds.")
    return result


ERC721_ABI = """
[
  {
    "inputs": [],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "constant": true,
    "inputs": [
      {
        "internalType": "address",
        "name": "owner",
        "type": "address"
      }
    ],
    "name": "balanceOf",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "internalType": "uint256",
        "name": "tokenId",
        "type": "uint256"
      }
    ],
    "name": "ownerOf",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "internalType": "address",
        "name": "from",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "tokenId",
        "type": "uint256"
      },
      {
        "internalType": "bytes",
        "name": "_data",
        "type": "bytes"
      }
    ],
    "name": "safeTransferFrom",
    "outputs": [],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "internalType": "address",
        "name": "from",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "tokenId",
        "type": "uint256"
      }
    ],
    "name": "safeTransferFrom",
    "outputs": [],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "internalType": "address",
        "name": "from",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "tokenId",
        "type": "uint256"
      }
    ],
    "name": "transferFrom",
    "outputs": [],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "tokenId",
        "type": "uint256"
      }
    ],
    "name": "approve",
    "outputs": [],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "approved",
        "type": "bool"
      }
    ],
    "name": "setApprovalForAll",
    "outputs": [],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "internalType": "uint256",
        "name": "tokenId",
        "type": "uint256"
      }
    ],
    "name": "getApproved",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "internalType": "address",
        "name": "owner",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "operator",
        "type": "address"
      }
    ],
    "name": "isApprovedForAll",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "internalType": "bytes4",
        "name": "interfaceId",
        "type": "bytes4"
      }
    ],
    "name": "supportsInterface",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "from",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "uint256",
        "name": "tokenId",
        "type": "uint256"
      }
    ],
    "name": "Transfer",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "owner",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "approved",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "uint256",
        "name": "tokenId",
        "type": "uint256"
      }
    ],
    "name": "Approval",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "owner",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "operator",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "approved",
        "type": "bool"
      }
    ],
    "name": "ApprovalForAll",
    "type": "event"
  }
]

"""
ERC1155_ABI = """
[
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "account",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "operator",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "approved",
        "type": "bool"
      }
    ],
    "name": "ApprovalForAll",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "operator",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "from",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256[]",
        "name": "ids",
        "type": "uint256[]"
      },
      {
        "indexed": false,
        "internalType": "uint256[]",
        "name": "values",
        "type": "uint256[]"
      }
    ],
    "name": "TransferBatch",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "operator",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "from",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "id",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "value",
        "type": "uint256"
      }
    ],
    "name": "TransferSingle",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "string",
        "name": "value",
        "type": "string"
      },
      {
        "indexed": true,
        "internalType": "uint256",
        "name": "id",
        "type": "uint256"
      }
    ],
    "name": "URI",
    "type": "event"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "id",
        "type": "uint256"
      }
    ],
    "name": "balanceOf",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address[]",
        "name": "accounts",
        "type": "address[]"
      },
      {
        "internalType": "uint256[]",
        "name": "ids",
        "type": "uint256[]"
      }
    ],
    "name": "balanceOfBatch",
    "outputs": [
      {
        "internalType": "uint256[]",
        "name": "",
        "type": "uint256[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "operator",
        "type": "address"
      }
    ],
    "name": "isApprovedForAll",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "from",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "internalType": "uint256[]",
        "name": "ids",
        "type": "uint256[]"
      },
      {
        "internalType": "uint256[]",
        "name": "amounts",
        "type": "uint256[]"
      },
      {
        "internalType": "bytes",
        "name": "data",
        "type": "bytes"
      }
    ],
    "name": "safeBatchTransferFrom",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "from",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "id",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      },
      {
        "internalType": "bytes",
        "name": "data",
        "type": "bytes"
      }
    ],
    "name": "safeTransferFrom",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "operator",
        "type": "address"
      },
      {
        "internalType": "bool",
        "name": "approved",
        "type": "bool"
      }
    ],
    "name": "setApprovalForAll",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes4",
        "name": "interfaceId",
        "type": "bytes4"
      }
    ],
    "name": "supportsInterface",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "id",
        "type": "uint256"
      }
    ],
    "name": "uri",
    "outputs": [
      {
        "internalType": "string",
        "name": "",
        "type": "string"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  }
]
"""



# This is an example of interacting with an ERC-20 contract, using the CustomToken interface.
import os
from fireblocks_defi_sdk_py import Chain, CustomToken, Web3Bridge, fetch_abi
from fireblocks_sdk import FireblocksSDK, TRANSACTION_STATUS_COMPLETED

API_KEY = os.environ['TEST_API_KEY']
API_SECRET = os.environ['SECRET']

SDK = FireblocksSDK(API_SECRET, API_KEY)
VAULT_ID = "0"
CONTRACT_ADDRESS = ""
RECEIVER_ADDRESS = ""
TOKEN_AMOUNT = 10

if __name__ == "__main__":
    # We start by building a bridge to the Ropsten network, including initiating a CustomToken object which will
    # represent our ERC20 contract. In order to initiate a CustomToken, we will need a contract ABI. This can be read
    # from another file, passed as a variable declared locally, or fetched from EtherScan using fetch_abi(address). We
    # will use the latter.
    ropsten_bridge = Web3Bridge(SDK, VAULT_ID, CONTRACT_ADDRESS, Chain.ROPSTEN)
    contract_abi = fetch_abi(CONTRACT_ADDRESS)
    erc20_contract_bridge = CustomToken(ropsten_bridge, contract_abi)

    # We build functions through passing the function name exactly as it appears on the ABI, following by its arguments.
    # It's important to differentiate between a read (view) and write function. We will demonstrate both.
    total_supply = erc20_contract_bridge.call_read_function("totalSupply")
    my_balance = erc20_contract_bridge.call_read_function("balanceOf", erc20_contract_bridge.wallet_address)

    # Note how both of the functions above were not submitted to Fireblocks, as there is no need to sign these (read).
    # The below functions will demonstrate a write function, that will afterwards be submitted to Fireblocks.

    # We will mint a token (function receives address & amount) to our own address and then send it to another address.
    # We will also have to add the sender under the building_params in the following format:
    checked_address = erc20_contract_bridge.web_provider.toChecksumAddress(erc20_contract_bridge.wallet_address)
    # Web3 only accepts checkedSum addresses.
    building_params = {"from": checked_address}
    mint_raw_transaction = erc20_contract_bridge.call_write_function("mint", checked_address,
                                                                     TOKEN_AMOUNT,
                                                                     building_params=building_params)
    # We will now submit the transaction to Fireblocks.
    mint_transaction = erc20_contract_bridge.submit_transaction(mint_raw_transaction)
    if ropsten_bridge.check_tx_is_completed(mint_transaction['id']) == TRANSACTION_STATUS_COMPLETED:
        print(f"Successfully minted {TOKEN_AMOUNT} to {erc20_contract_bridge.wallet_address}")
        transfer_raw_transaction = erc20_contract_bridge.call_write_function("transferFrom",
                                                                             checked_address,
                                                                             RECEIVER_ADDRESS, TOKEN_AMOUNT,
                                                                             building_params=building_params)
        transfer_transaction = erc20_contract_bridge.submit_transaction(transfer_raw_transaction,
                                                                        "Transferring 10 minted tokens.")
        if ropsten_bridge.check_tx_is_completed(transfer_transaction['id']):
            print(
                f"Successfully transferred {TOKEN_AMOUNT} from {erc20_contract_bridge.wallet_address} to "
                f"{RECEIVER_ADDRESS}")
        else:
            print("Failed transferring minted tokens.")
    else:
        print("Failed minting new tokens.")



from . base_token import *
from .. utils.helpers import ERC1155_ABI


class ERC1155(BaseToken):
    def __init__(self, web3_bridge: Web3Bridge):
        super().__init__(web3_bridge)
        self.abi = ERC1155_ABI
        self.contract: contract = self.web_provider.eth.contract(
            address=self.web_provider.toChecksumAddress(self.web3_bridge.external_wallet_address),
            abi=self.abi
        )

    # Payables
    def set_approval_for_all(self, operator_address: str, is_approved: bool, note: str = ""):
        """
        Provide an operator with approval permission or revoke them
        :param operator_address: Approved operator address
        :param is_approved: True to permit, False to revoke
        :param note: (Optional) Add a note to the transaction.
        :return: None
        """
        checked_op_adr = self.web_provider.toChecksumAddress(operator_address)
        return self.submit_transaction(self.call_write_function("setApprovalForAll", checked_op_adr, is_approved), note)

    def safe_transfer_from(self, to_address: str, token_id: int, amount: int, from_address: str = "",
                           data: bytes = bytearray(), note: str = ""):
        """
        :param to_address: The receiver of the token.
        :param token_id: The ID of the sent token.
        :param amount: Amount of Token ID to send.
        :param from_address: (Optional) The sender of the token, if it's not the wallet under Fireblocks.
        :param data: (Optional) Send additional data (bytes) only if required by contract.
        :param note: (Optional) Add a note to the transaction.
        :return: None
        """
        if not from_address:
            from_address = self.wallet_address
        address_dict = {"from": self.web_provider.toChecksumAddress(from_address)}
        checked_from_adr = self.web_provider.toChecksumAddress(from_address)
        checked_to_adr = self.web_provider.toChecksumAddress(to_address)

        transaction = self.contract.functions.safeTransferFrom(
            checked_from_adr,
            checked_to_adr,
            token_id,
            amount,
            data
        ).buildTransaction(address_dict)

        return self.submit_transaction(transaction, note)

    def safe_batch_transfer_from(self, to_address: str, token_ids: list[int], values: list[int], from_address: str = "",
                                 data: bytes = bytearray(), note: str = ""):
        """
        Length of token_ids and values must match. Moreover, the value of each token (at position x at [values]) to be
        sent must match the same position at [token_ids] (position x). For example:
        token_ids [3, 6, 10] and values [10, 20, 30] will send:
        - 10 tokens of token id 3
        - 20 tokens of token id 6
        - 30 tokens of token id 10
        :param to_address: The receiver of the token.
        :param token_ids: A list of token ids to transfer from sender to receiver.
        :param values: A list of values to transfer of each token.
        :param from_address: (Optional) The sender of the token, if it's not the wallet under Fireblocks.
        :param data: (Optional) Send additional data (bytes) only if required by contract.
        :param note: (Optional) Add a note to the transaction.
        :return: None
        """
        if not from_address:
            from_address = self.wallet_address
        address_dict = {"from": self.web_provider.toChecksumAddress(from_address)}
        checked_from_adr = self.web_provider.toChecksumAddress(from_address)
        checked_to_adr = self.web_provider.toChecksumAddress(to_address)
        if len(token_ids) != len(values):
            raise ValueError("Length of token_ids and values must match!")

        transaction = self.contract.functions.safeBatchTransferFrom(
            checked_from_adr,
            checked_to_adr,
            token_ids,
            values,
            data
        ).buildTransaction(address_dict)

        return self.submit_transaction(transaction, note)

    # Views
    def supports_interface(self, interface_id: str = "0xd9b67a26") -> bool:
        """
        Checks if contract supports a certain interface.
        :param interface_id: (Optional) The interface id. "0x80ac58cd" is ERC721 interface id. "0xd9b67a26" is ERC1155
         interface id. Checks class interface id by default.
        :return: True if the contract supports the interface, False otherwise
        """
        return self.call_read_function("supportsInterface", interface_id)

    def balance_of(self, token_id: int, owner_address: str = "") -> int:
        """
        Gets the balance of the address provided
        :param owner_address: Address to be checked
        :param token_id: ID of the token.
        :return: Balance (int)
        """
        if not owner_address:
            owner_address = self.wallet_address
        return self.call_read_function("balanceOf", owner_address, token_id)

    def balance_of_batch(self, id_list: list[int], owners_list=None) -> list[int]:
        """

        :param owners_list: A list of addresses
        :param id_list: A list of token Ids
        :return:
        """
        if not owners_list:
            owners_list = [self.wallet_address] * len(id_list)
        checked_addresses = [self.web_provider.toChecksumAddress(address) for address in owners_list]
        return self.call_read_function("balanceOfBatch", checked_addresses, id_list)

    def is_approved_for_all(self, operator_address: str, owner_address: str = "") -> bool:
        """
        Validates whether an operator has been approved by a stated owner.
        :param operator_address: Address of the meant to be operator.
        :param owner_address: (Optional) Address of the owner, in case it's not the wallet address.
        :return: True whether operator is approved, False otherwise
        """
        if not owner_address:
            owner_address = self.wallet_address
        owner_checked_address = self.web_provider.toChecksumAddress(owner_address)
        operator_checked_address = self.web_provider.toChecksumAddress(operator_address)
        return self.call_read_function("isApprovedForAll", owner_checked_address, operator_checked_address)

    def uri(self, token_id: int) -> str:
        """
        :param token_id: URI id of a contract
        :return: The JSON metadata
        """
        return self.call_read_function("uri", token_id)

    # Events



from distutils.core import setup

setup(
    name='lv-test-defi',
    packages=['lv-test-defi',
              'lv-test-defi.tokenization',
              'lv-test-defi.tokenization.tokens',
              'lv-test-defi.tokenization.utils',
              'lv-test-defi.tokenization.examples'],
    version='1.2',
    license='MIT',
    description='lv-test-defi',
    author='Fireblocks',
    author_email='fireblocks@fireblocks.com',
    url='https://github.com/fireblocks/fireblocks-defi-sdk-py',
    download_url='https://github.com/fireblocks/fireblocks-defi-sdk-py/archive/refs/tags/1.1.tar.gz',
    keywords=['FIREBLOCKS', 'DeFi', 'SDK', 'PYTHON'],

    install_requires=[
        'fireblocks_sdk',
        'web3',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)



# This is an example of interacting with a custom contract, using the CustomToken interface.
import os
from fireblocks_defi_sdk_py import Chain, CustomToken, Web3Bridge
from fireblocks_sdk import FireblocksSDK, TRANSACTION_STATUS_COMPLETED

API_KEY = os.environ['TEST_API_KEY']
API_SECRET = os.environ['SECRET']

SDK = FireblocksSDK(API_SECRET, API_KEY)
VAULT_ID, SECONDARY_VAULT_ID = "0", "2"
CONTRACT_ADDRESS = ""
WHITELISTED_CONTRACT_UUID = ""  # This is the Fireblocks UUID of the whitelisted address.

FILE_PATH, MODE = "some/path/to/my/contract", "r"

if __name__ == "__main__":
    # We start by building a bridge to the Ropsten network, including initiating a CustomToken object which will
    # represent our custom contract. In order to initiate a CustomToken, we will need a contract ABI. This can be read
    # from another file, passed as a variable declared locally, or fetched from EtherScan using fetch_abi(address). We
    # will use the first.
    ropsten_bridge = Web3Bridge(SDK, VAULT_ID, CONTRACT_ADDRESS, Chain.ROPSTEN, WHITELISTED_CONTRACT_UUID)
    with open(FILE_PATH, MODE) as file:
        contract_abi = file.read()
    custom_contract_bridge = CustomToken(ropsten_bridge, contract_abi)

    # Let's assume our ABI holds a transferOwnership write function, and a contractOwner function (returning a boolean).
    # We will first validate we own the contract and then transfer it to another vault in our account.
    checked_address = custom_contract_bridge.web_provider.toChecksumAddress(custom_contract_bridge.wallet_address)
    if custom_contract_bridge.call_read_function("contractOwner", checked_address):
        # We will initiate another bridge, as we will need the wallet address, and we will later on transfer it back.
        secondary_ropsten_bridge = Web3Bridge(SDK, SECONDARY_VAULT_ID, CONTRACT_ADDRESS, Chain.ROPSTEN,
                                              WHITELISTED_CONTRACT_UUID)
        secondary_contract_bridge = CustomToken(secondary_ropsten_bridge, contract_abi)
        building_params = {"from": checked_address}
        secondary_address = secondary_contract_bridge.web_provider.toChecksumAddress(
            secondary_contract_bridge.wallet_address)
        transfer_ownership_raw_transaction = custom_contract_bridge. \
            call_write_function("transferOwnership",
                                secondary_address,
                                building_params=building_params)

        transfer_ownership_transaction = custom_contract_bridge.submit_transaction(transfer_ownership_raw_transaction)
        if ropsten_bridge.check_tx_is_completed(transfer_ownership_transaction['id']) == TRANSACTION_STATUS_COMPLETED:
            print("Successfully transferred ownership from first wallet to the second one.")
            # We will now transfer the ownership back to our original wallet.
            building_params['from'] = secondary_address
            ownership_raw_transaction = secondary_contract_bridge. \
                call_write_function("transferOwnership",
                                    checked_address,
                                    building_params=building_params)
            ownership_transaction = secondary_contract_bridge.submit_transaction(ownership_raw_transaction, "Returning")
            if secondary_ropsten_bridge.check_tx_is_completed(ownership_transaction['id']) == \
                    TRANSACTION_STATUS_COMPLETED:
                print("Successfully transferred ownership from second wallet to the first one.")
            else:
                print("Failed transferring ownership from the second wallet to the first.")
        else:
            print("Failed transferring ownership from the first wallet to the second.")
